@@ -3,6 +3,10 @@
 {-# LANGUAGE NoMonomorphismRestriction #-}
 module Language.Haskell.TH.ExpandSyns(-- * Expand synonyms
                                       expandSyns
+                                     ,expandSynsWith
+                                     ,SynonymExpansionSettings
+                                     ,noWarnTypeFamilies
+
                                       -- * Misc utilities
                                      ,substInType
                                      ,substInCon
@@ -12,6 +16,8 @@ import Language.Haskell.TH hiding(cxt)
 import qualified Data.Set as Set
 import Data.Generics
 import Control.Monad
+import Data.Monoid
+import Prelude
 
 -- For ghci
 #ifndef MIN_VERSION_template_haskell
@@ -65,28 +71,32 @@ tyVarBndrSetName n _ = n
 (<*>) :: (Monad m) => m (a -> b) -> m a -> m b
 (<*>) = ap
 
-type SynInfo = ([Name],Type)
-
-nameIsSyn :: Name -> Q (Maybe SynInfo)
-nameIsSyn n = do
-  i <- reify n
-  case i of
-    TyConI d -> decIsSyn d
-    ClassI {} -> return Nothing
-    PrimTyConI {} -> return Nothing
-#if MIN_VERSION_template_haskell(2,11,0)
-    FamilyI (OpenTypeFamilyD (TypeFamilyHead name _ _ _)) _ -> maybeWarnTypeFamily TypeFam name >> return Nothing
-    FamilyI (ClosedTypeFamilyD (TypeFamilyHead name _ _ _) _) _ -> maybeWarnTypeFamily TypeFam name >> return Nothing
-    FamilyI (DataFamilyD _ _ _) _ -> return Nothing
-#elif MIN_VERSION_template_haskell(2,7,0)
-    FamilyI (FamilyD flavour name _ _) _ -> maybeWarnTypeFamily flavour name >> return Nothing
-#endif
-    _ -> do
-            warn ("Don't know how to interpret the result of reify "++show n++" (= "++show i++").\n"++
-                  "I will assume that "++show n++" is not a type synonym.")
-            return Nothing
 
 
+data SynonymExpansionSettings =
+  SynonymExpansionSettings {
+    sesWarnTypeFamilies :: Bool
+  }
+
+
+
+-- | Default settings ('mempty'):
+--
+-- * Warn if type families are encountered.
+--
+-- (The 'mappend' is currently rather useless; the monoid instance is intended for additional settings in the future).
+instance Monoid SynonymExpansionSettings where
+  mempty =
+    SynonymExpansionSettings {
+      sesWarnTypeFamilies = True
+    }
+
+  mappend (SynonymExpansionSettings w1) (SynonymExpansionSettings w2) =
+    SynonymExpansionSettings (w1 && w2)
+
+-- | Suppresses the warning that type families are unsupported.
+noWarnTypeFamilies :: SynonymExpansionSettings
+noWarnTypeFamilies = mempty { sesWarnTypeFamilies = False }
 
 warn ::  String -> Q ()
 warn msg =
@@ -95,50 +105,125 @@ warn msg =
 #else
     report False
 #endif
-      (packagename ++": "++"WARNING: "++msg)
+      (packagename ++": WARNING: "++msg)
 
 
+
+
+type SynInfo = ([Name],Type)
+
+nameIsSyn :: SynonymExpansionSettings -> Name -> Q (Maybe SynInfo)
+nameIsSyn settings n = do
+  i <- reify n
+  case i of
+    ClassI {} -> no
+    ClassOpI {} -> no
+    TyConI d -> decIsSyn settings d
+#if MIN_VERSION_template_haskell(2,7,0)
+    FamilyI d _ -> decIsSyn settings d -- Called for warnings
+#endif
+    PrimTyConI {} -> no
+    DataConI {} -> no
+    VarI {} -> no
+    TyVarI {} -> no
+#if MIN_VERSION_template_haskell(2,12,0)
+    PatSynI {} -> no
+#endif
+
+  where
+    no = return Nothing
+
+decIsSyn :: SynonymExpansionSettings -> Dec -> Q (Maybe SynInfo)
+decIsSyn settings = go
+  where
+    go (TySynD _ vars t) = return (Just (tyVarBndrGetName <$> vars,t))
+
+#if MIN_VERSION_template_haskell(2,11,0)
+    go (OpenTypeFamilyD (TypeFamilyHead name _ _ _)) = maybeWarnTypeFamily settings name >> no
+    go (ClosedTypeFamilyD (TypeFamilyHead name _ _ _) _) = maybeWarnTypeFamily settings name >> no
+#else
+
+#if MIN_VERSION_template_haskell(2,9,0)
+    go (ClosedTypeFamilyD name _ _ _) = maybeWarnTypeFamily settings name >> no
+#endif
+
+    go (FamilyD TypeFam name _ _) = maybeWarnTypeFamily settings name >> no
+#endif
+
+    go (FunD {}) = no
+    go (ValD {}) = no
+    go (DataD {}) = no
+    go (NewtypeD {}) = no
+    go (ClassD {}) = no
+    go (InstanceD {}) = no
+    go (SigD {}) = no
+    go (ForeignD {}) = no
+
+#if MIN_VERSION_template_haskell(2,8,0)
+    go (InfixD {}) = no
+#endif
+
 #if MIN_VERSION_template_haskell(2,4,0)
-maybeWarnTypeFamily :: FamFlavour -> Name -> Q ()
-maybeWarnTypeFamily flavour name =
-  case flavour of
-    TypeFam ->
-      warn ("Type synonym families (and associated type synonyms) are currently not supported (they won't be expanded). Name of unsupported family: "++show name)
-
-    DataFam -> return ()
-      -- Nothing to expand for data families, so no warning
+    go (PragmaD {}) = no
 #endif
 
--- | Handles only declaration constructs that can be returned by 'reify'ing a type name.
-decIsSyn :: Dec -> Q (Maybe SynInfo)
-decIsSyn (ClassD {}) = return Nothing
-decIsSyn (DataD {}) = return Nothing
-decIsSyn (NewtypeD {}) = return Nothing
-decIsSyn (TySynD _ vars t) = return (Just (tyVarBndrGetName <$> vars,t))
+    -- Nothing to expand for data families, so no warning
 #if MIN_VERSION_template_haskell(2,11,0)
-decIsSyn (OpenTypeFamilyD (TypeFamilyHead name _ _ _)) = maybeWarnTypeFamily TypeFam name >> return Nothing
-decIsSyn (ClosedTypeFamilyD (TypeFamilyHead name _ _ _) _) = maybeWarnTypeFamily TypeFam name >> return Nothing
-decIsSyn (DataFamilyD _ _ _) = return Nothing
+    go (DataFamilyD {}) = no
 #elif MIN_VERSION_template_haskell(2,4,0)
-decIsSyn (FamilyD flavour name _ _) = maybeWarnTypeFamily flavour name >> return Nothing
+    go (FamilyD DataFam _ _ _) = no
+#endif
+
+#if MIN_VERSION_template_haskell(2,4,0)
+    go (DataInstD {}) = no
+    go (NewtypeInstD {}) = no
+    go (TySynInstD {}) = no
+#endif
+
+#if MIN_VERSION_template_haskell(2,9,0)
+    go (RoleAnnotD {}) = no
+#endif
+
+#if MIN_VERSION_template_haskell(2,10,0)
+    go (StandaloneDerivD {}) = no
+    go (DefaultSigD {}) = no
+#endif
+
+#if MIN_VERSION_template_haskell(2,12,0)
+    go (PatSynD {}) = no
+    go (PatSynSigD {}) = no
+#endif
+
+    no = return Nothing
+
+#if MIN_VERSION_template_haskell(2,4,0)
+maybeWarnTypeFamily :: SynonymExpansionSettings -> Name -> Q ()
+maybeWarnTypeFamily settings name =
+  when (sesWarnTypeFamilies settings) $
+      warn ("Type synonym families (and associated type synonyms) are currently not supported (they won't be expanded). Name of unsupported family: "++show name)
 #endif
-decIsSyn x = do
-    warn ("Unrecognized declaration construct: "++ show x++". I will assume that it's not a type synonym declaration.")
-    return Nothing
 
 
 
 
 
+
+
+-- | Calls 'expandSynsWith' with the default settings.
+expandSyns :: Type -> Q Type
+expandSyns = expandSynsWith mempty
+
+
 -- | Expands all type synonyms in the given type. Type families currently won't be expanded (but will be passed through).
-expandSyns :: Type -> Q Type
-expandSyns = \t ->
+expandSynsWith :: SynonymExpansionSettings -> Type -> Q Type
+expandSynsWith settings = expandSyns'
+
+    where
+      expandSyns' t =
          do
            (acc,t') <- go [] t
            return (foldl AppT t' acc)
 
-
-    where
       -- Must only be called on an `x' requiring no expansion
       passThrough acc x = return (acc, x)
 
@@ -158,8 +243,8 @@ expandSyns = \t ->
       go acc x@(VarT _) = passThrough acc x
 
       go [] (ForallT ns cxt t) = do
-        cxt' <- mapM (bindPred expandSyns) cxt
-        t' <- expandSyns t
+        cxt' <- mapM (bindPred expandSyns') cxt
+        t' <- expandSyns' t
         return ([], ForallT ns cxt' t')
 
       go acc x@(ForallT _ _ _) =
@@ -169,17 +254,17 @@ expandSyns = \t ->
 
       go acc (AppT t1 t2) =
           do
-            r <- expandSyns t2
+            r <- expandSyns' t2
             go (r:acc) t1
 
       go acc x@(ConT n) =
           do
-            i <- nameIsSyn n
+            i <- nameIsSyn settings n
             case i of
               Nothing -> return (acc, x)
               Just (vars,body) ->
                   if length acc < length vars
-                  then fail (packagename++": expandSyns: Underapplied type synonym: "++show(n,acc))
+                  then fail (packagename++": expandSynsWith: Underapplied type synonym: "++show(n,acc))
                   else
                       let
                           substs = zip vars acc
@@ -220,13 +305,13 @@ expandSyns = \t ->
 #if MIN_VERSION_template_haskell(2,11,0)
       go acc (InfixT t1 nm t2) =
           do
-            t1' <- expandSyns t1
-            t2' <- expandSyns t2
+            t1' <- expandSyns' t1
+            t2' <- expandSyns' t2
             return (acc,InfixT t1' nm t2')
       go acc (UInfixT t1 nm t2) =
           do
-            t1' <- expandSyns t1
-            t2' <- expandSyns t2
+            t1' <- expandSyns' t1
+            t2' <- expandSyns' t2
             return (acc,UInfixT t1' nm t2')
       go acc (ParensT t) =
           do
@@ -235,6 +320,10 @@ expandSyns = \t ->
       go acc x@WildCardT = passThrough acc x
 #endif
 
+#if MIN_VERSION_template_haskell(2,12,0)
+      go acc x@(UnboxedSumT _) = passThrough acc x
+#endif
+
 class SubstTypeVariable a where
     -- | Capture-free substitution
     subst :: (Name, Type) -> a -> a
@@ -284,6 +373,10 @@ instance SubstTypeVariable Type where
       go s@WildCardT = s
 #endif
 
+#if MIN_VERSION_template_haskell(2,12,0)
+      go s@(UnboxedSumT _) = s
+#endif
+
 -- testCapture :: Type
 -- testCapture =
 --     let
